open Ur

fun isIdent1 ch = Char.isLower ch || ch = #"_"
fun isIdent ch = Char.isUpper ch || Char.isLower ch || Char.isDigit ch || ch = #"_"

fun cident s =
    if String.length s > 0 && Char.isUpper (String.sub s 0) && String.all isIdent s then
        s
    else
        error <xml>Invalid cident: {[s]}</xml>

fun ident s =
    if String.length s > 0 && isIdent1 (String.sub s 0) && String.all isIdent s then
        s
    else
        error <xml>Invalid ident: {[s]}</xml>

fun ppF c =
    case c of
        CVar ([], x) => ident x
      | CName s =>
        if String.length s > 0 && String.all Char.isDigit s then
            s
        else
            cident s
      | _ => error <xml>Invalid Name</xml>

fun ppV' f (xs, x) = List.foldr (fn x s => cident x ^ "." ^ s) (f x) xs

val ppV = ppV' ident
val ppM = ppV' cident

fun ppC c =
    case c of
        TFun (dom, ran) => "(" ^ ppC dom ^ " -> " ^ ppC ran ^ ")"
      | TRecord c => "$" ^ ppC c
      | CVar v => ppV v
      | CApp (f, x) => "(" ^ ppC f ^ " " ^ ppC x ^ ")"
      | CName s => "#" ^ cident s
      | CRecord [] => "[]"
      | CRecord ((x, c) :: []) => "[" ^ ppF x ^ " = " ^ ppC c ^ "]"
      | CRecord ((x, c) :: xcs) => "[" ^ ppF x ^ " = " ^ ppC c
                                   ^ List.foldl (fn (x, c) s =>
                                                    s ^ ", " ^ ppF x ^ " = " ^ ppC c)
                                                "]" xcs
      | CUnit => "()"
      | CTuple [] => error <xml>Empty CTuple</xml>
      | CTuple (_ :: []) => error <xml>Singleton CTuple</xml>
      | CTuple (c :: cs) => "(" ^ ppC c
                            ^ List.foldl (fn c s =>
                                             s ^ ", " ^ ppC c) ")" cs

fun ppP p =
    case p of
        Int n => show n
      | Float n => show n
      | String s => "\"" ^ GraftidFfi.escape s ^ "\""
      | Char c => "#\"" ^ GraftidFfi.escape (String.str c) ^ "\""

fun ppE e =
    case e of
        EPrim p => ppP p
      | EVar v => ppV v
      | EApp (e1, e2) => "(" ^ ppE e1 ^ " " ^ ppE e2 ^ ")"
      | EAbs (x, e) => "(fn " ^ ident x ^ " => " ^ ppE e ^ ")"
      | ECApp (e, c) => "(" ^ ppE e ^ " [" ^ ppC c ^ "])"
      | EDisjointApp e => "(" ^ ppE e ^ " !)"
      | ERecord [] => "{}"
      | ERecord ((x, e) :: []) => "{" ^ ppF x ^ " = " ^ ppE e ^ "}"
      | ERecord ((x, e) :: xes) => "{" ^ ppF x ^ " = " ^ ppE e
                                   ^ List.foldl (fn (x, e) s =>
                                                    s ^ ", " ^ ppF x ^ " = " ^ ppE e)
                                                "" xes ^ "}"
      | EField (e, f) => "(" ^ ppE e ^ "." ^ ppF f ^ ")"
      | ELet (x, e1, e2) => "(let val " ^ ident x ^ " = " ^ ppE e1 ^ " in " ^ ppE e2 ^ " end)"

fun ppD d =
    case d of
        DCon (x, c) => "con " ^ ident x ^ " = " ^ ppC c
      | DVal (x, e) => "val " ^ ident x ^ " = " ^ ppE e
      | DValRec [] => error <xml>Empty DValRec</xml>
      | DValRec ((x, e) :: []) => "val rec " ^ ident x ^ " = " ^ ppE e
      | DValRec ((x, e) :: xes) => "val rec " ^ ident x ^ " = " ^ ppE e
                                   ^ List.foldl (fn (x, e) s =>
                                                    s ^ " and " ^ ident x ^ " = " ^ ppE e) "" xes
      | DStr (x, s) => "structure " ^ cident x ^ " = " ^ ppS s
      | DOpen v => "open " ^ ppM v
      | DOpenApp (ms, m, s) => "open " ^ ppM (ms, m) ^ "(" ^ ppS s ^ ")"

and ppS s =
    case s of
        StrConst ds => List.foldl (fn d s => s ^ "\t" ^ ppD d ^ "\n") "struct\n" ds ^ "end"
      | StrVar x => cident x
      | StrProj (s, x) => ppS s ^ "." ^ cident x
      | StrApp (s1, s2) => ppS s1 ^ "(" ^ ppS s2 ^ ")"

val pp = List.foldl (fn d s => s ^ ppD d ^ "\n") ""
